#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <pthread.h>
#include <sys/time.h>

#include "cspmain.h"
#include "cspcom.h"


///////////////////////////////////////////////////////////////////////////////

unsigned int CSP::GetTickCount()
{
    struct timeval tv;
    gettimeofday( &tv, NULL );
    // this will rollover ~ every 49.7 days
    return (unsigned int)( (tv.tv_sec-startime.tv_sec) * 1000 + tv.tv_usec / 1000);
}


///////////////////////////////////////////////////////////////////////////////
// READ CFG
///////////////////////////////////////////////////////////////////////////////

int CSP::read_config(struct config_data *cfg)
{
	char currentline[1024];
	char str[255];
	FILE *fhandle;

	cspwait=3000;
	// Open Config file
	fhandle = fopen(config_file,"r");
	if (fhandle==0) {
		printf("Cannot open file '%s'\n",config_file);
		return -1;
	} 

	//printf(" config: parsing file '%s'\n",config_file);

	// Init config data
	memset( cfg, 0, sizeof(struct config_data) );

	while (!feof(fhandle)) {
		if ( !fgets(currentline, sizeof(currentline)-1 , fhandle) ) break;
		
		// Debug
		if (memcmp(currentline,"DEBUG:",6)==0)
			{
			sscanf (currentline,"DEBUG: %d",&dbg);
			}
		//cacheport
		if (memcmp(currentline,"PORT:",5)==0)
			{
			sscanf (currentline,"PORT: %s",str);
			cfg->cache.port = atoi(str);
			}
		//cache peer
		if (memcmp(currentline,"PEER:",5)==0)
				{
				// TODO: check for reuse of same peer
				struct cachepeer_data *peer = (struct cachepeer_data *)malloc( sizeof(struct cachepeer_data) );
				memset(peer,0,sizeof(struct cachepeer_data) );
				sscanf (currentline,"PEER: %s %s",peer->host,str);
				peer->port = atoi(str);
				peer->ip = hostname2ip(peer->host);
				peer->checkiptime = 0;
				peer->reqnb=0;
				peer->id=cfg->cachepeerid; cfg->cachepeerid++;
				peer->next = cfg->cache.peer;
				cfg->cache.peer = peer;
				}
		// CSP wait
		if (memcmp(currentline,"WAIT:",5)==0)
			{
			sscanf (currentline,"WAIT: %d",&cspwait);
			}

		}

	fclose(fhandle);
	return 0;
}


struct cache_data * CSP::cache_add( struct cache_data *newdata )
{
	int i = icachetab;
	memset( &cachetab[i], 0, sizeof(struct cache_data) );
	cachetab[i].status = CACHE_STAT_WAIT; // 0:Wait; 1: dcw received
	cachetab[i].recvtime = GetTickCount();
	cachetab[i].tag = newdata->tag;
	cachetab[i].sid = newdata->sid;
	cachetab[i].onid = newdata->onid;
	cachetab[i].caid = newdata->caid;
	cachetab[i].hash = newdata->hash;
	cachetab[i].prov = newdata->prov;
	icachetab++;
	if (icachetab>=MAX_CACHETAB) icachetab=0;
	return &cachetab[i];
}

struct cache_data * CSP::cache_fetch( struct cache_data *thereq )
{
	//printf("Query CSP Cache -> %d (%02x:%04x:%04x:%08x)\n",MAX_CACHETAB,thereq->tag,thereq->caid,thereq->sid,thereq->hash);
	for(int i=0; i<MAX_CACHETAB; i++) 
		{
		if (cachetab[i].status!=CACHE_STAT_DEL)  //printf("fetch X:%d is done\n",i);
			if ( (cachetab[i].hash==thereq->hash)&&(cachetab[i].caid==thereq->caid)&&(cachetab[i].tag==thereq->tag)&&(cachetab[i].sid==thereq->sid) )
					return &cachetab[i];
		}
	//printf("fetch 0 is done\n");
	return NULL;
}

///////////////////////////////////////////////////////////////////////////////
// SEND DATA
///////////////////////////////////////////////////////////////////////////////

void CSP::sendtoip( unsigned int ip, int port, unsigned char *buf, int len)
{
	if (ip && port) {
		struct sockaddr_in si_other;
		int slen=sizeof(si_other);
		memset((char *) &si_other, 0, sizeof(si_other));
		si_other.sin_family = AF_INET;
		si_other.sin_port = htons( port );
		si_other.sin_addr.s_addr = ip;

		while (1) {
			int retval = sendto(outsock, buf, len, MSG_NOSIGNAL, (struct sockaddr *)&si_other, slen);
			if (retval>0) break;
			if ( (retval==-1)&&(errno!=EINTR)&&(errno!=EAGAIN)&&(errno!=EWOULDBLOCK) ) break;
		}
	}
}

void CSP::send_request(struct cache_data *pcache,struct cachepeer_data *peer)
{
	unsigned char buf[16];
	buf[0] = TYPE_REQUEST;
	buf[1] = pcache->tag;
	buf[2] = pcache->sid>>8;
	buf[3] = pcache->sid;
	buf[4] = pcache->onid>>8;
	buf[5] = pcache->onid&0xff;
	buf[6] = pcache->caid>>8;
	buf[7] = pcache->caid;
	buf[8] = pcache->hash>>24;
	buf[9] = pcache->hash>>16;
	buf[10] = pcache->hash>>8;
	buf[11] = pcache->hash;
	sendtoip(peer->ip,peer->port, buf, 12);
	if (dbg) printf(" -> request to %s:%d (%04x:%04x:%08x)\n", peer->host, peer->port, pcache->caid, pcache->sid, pcache->hash);
}

void CSP::send_reply(struct cache_data *pcache,struct cachepeer_data *peer)
{
	unsigned char buf[64];
	//Common Data
	buf[0] = TYPE_REPLY;
	buf[1] = pcache->tag;
	buf[2] = pcache->sid>>8;
	buf[3] = pcache->sid;
	buf[4] = pcache->onid>>8;
	buf[5] = pcache->onid&0xff;
	buf[6] = pcache->caid>>8;
	buf[7] = pcache->caid;
	buf[8] = pcache->hash>>24;
	buf[9] = pcache->hash>>16;
	buf[10] = pcache->hash>>8;
	buf[11] = pcache->hash;
	buf[12] = pcache->tag;
	if (pcache->status==CACHE_STAT_WAIT)
		sendtoip(peer->ip,peer->port, buf, 13); // Failed reply
	else {
		memcpy( buf+13, pcache->cw, 16);
		sendtoip(peer->ip,peer->port, buf, 29);
		if (dbg) printf(" => reply to %s:%d (%04x:%04x:%08x)\n", peer->host, peer->port, pcache->caid, pcache->sid, pcache->hash);
	}
}

void CSP::send_pingreq(struct cachepeer_data *peer)
{
	unsigned char buf[32];
	buf[0] = TYPE_PINGREQ;
	// Ping data
	buf[1] = 0;
	buf[2] = 0;
	buf[3] = 0;
	buf[4] = 0;
	buf[5] = 0;
	buf[6] = 0;
	buf[7] = peer->id>>8; 
	buf[8] = peer->id&0xff;
	//Port
	buf[9] = 0;
	buf[10] = 0;
	buf[11] = cfg.cache.port>>8;
	buf[12] = cfg.cache.port&0xff;
	//Extra Data
	//Program Name
	buf[13] = 0x01; //ID
	buf[14] = 7; //LEN
	buf[15] = 'T'; buf[16] = 'r'; buf[17] = 'a'; buf[18] = 'c'; buf[19] = 'k'; buf[20] = 'e'; buf[21] = 'r';
	//Program Version
	buf[22] = 0x02; //ID
	buf[23] = 3; //LEN
	buf[24] = 'r'; buf[25] = '1';

	sendtoip(peer->ip,peer->port, buf, 26);
	if (dbg) printf(" > ping to %s:%d\n", peer->host, peer->port);
}


///////////////////////////////////////////////////////////////////////////////
// PEER
///////////////////////////////////////////////////////////////////////////////

struct cachepeer_data * CSP::getpeerbyip(unsigned int ip)
{
	struct cachepeer_data *peer = cfg.cache.peer;
	while(peer) {
		if (peer->ip==ip) return peer;
		peer = peer->next;
	}
	return NULL;
}

struct cachepeer_data * CSP::getpeerbyaddr(unsigned int ip, unsigned short port)
{
	struct cachepeer_data *peer = cfg.cache.peer;
	while(peer) {
		if ( (peer->ip==ip)&&(peer->recvport==port) ) return peer;
		peer = peer->next;
	}
	return NULL;
}

struct cachepeer_data * CSP::getpeerbyid(int id)
{
	struct cachepeer_data *peer = cfg.cache.peer;
	while(peer) {
		if (peer->id==id) return peer;
		peer = peer->next;
	}
	return NULL;
}


///////////////////////////////////////////////////////////////////////////////
// RECV MESSAGE
///////////////////////////////////////////////////////////////////////////////

void CSP::cache_recvmsg()
{
int port;
struct cachepeer_data *peerip;
struct cachepeer_data *peer;
struct cache_data req;
struct cache_data *pcache;
struct cachepeer_data *p;
int peerid;

	unsigned char buf[1024];
	struct sockaddr_in si_other;
	socklen_t slen=sizeof(si_other);
	int received = recvfrom( cfg.cache.handle, buf, sizeof(buf), 0, (struct sockaddr*)&si_other, &slen);

	unsigned int recv_ip;
	unsigned short recv_port;
	memcpy( &recv_ip, &si_other.sin_addr, 4);
	recv_port = ntohs(si_other.sin_port);

	if (received>0) {
		
		switch(buf[0]) {
			case TYPE_REQUEST:
				if (received<12) break;
				// Check Peer
				peer = getpeerbyaddr(recv_ip,recv_port);
				if (!peer) {
					peer = getpeerbyip(recv_ip);
					if (!peer) break;
				}
				peer->lastactivity = GetTickCount();
				//Get Cache Data
				if (received==20) { // only csp sends arbiter number
					strcpy(peer->program, "CSP");
					break;
				}
				peer->reqnb++;
				req.tag = buf[1];
				req.sid = (buf[2]<<8) | buf[3];
				req.onid = (buf[4]<<8) | buf[5];
				req.caid = (buf[6]<<8) | buf[7];
				req.hash = (buf[8]<<24) | (buf[9]<<16) | (buf[10]<<8) |buf[11];
				// Check for same request
				pcache = cache_fetch( &req );
				if (pcache==NULL) {
					pcache = cache_add( &req );
					if (dbg) printf(" <- request from %s:%d (%02x:%04x:%04x:%08x)\n", peer->host, peer->port, pcache->tag, pcache->caid, pcache->sid, pcache->hash);
					// Send to all Peers
					p = cfg.cache.peer;
					while (p) {
						if (p!=peer)
						if ( (peer->lastactivity+300000)>GetTickCount() ) // send only to active peers (5mn)
							send_request(pcache,p);
						p = p->next;
					}
				}
				break;

			case TYPE_REPLY:
				if (received<29) break;
				// Check Peer
				peer = getpeerbyaddr(recv_ip,recv_port);
				if (!peer) 
					{
					peer = getpeerbyip(recv_ip);
					if (!peer) break;
					}
				peer->lastactivity = GetTickCount();
				if (buf[12]!=buf[1]) break;
				req.tag = buf[1];
				req.sid = (buf[2]<<8) | buf[3];
				req.onid = (buf[4]<<8) | buf[5];
				req.caid = (buf[6]<<8) | buf[7];
				req.hash = (buf[8]<<24) | (buf[9]<<16) | (buf[10]<<8) |buf[11];
				
				if ( !acceptDCW(buf+13) ) break;
				peer->reqok++;
				// Search for data
				pcache = cache_fetch( &req );
				if (pcache==NULL) pcache = cache_add( &req );
				if (pcache->status!=CACHE_STAT_DCW) 
					{
					if (dbg) printf(" <= reply from %s:%d (%02x:%04x:%04x:%08x)\n", 
							peer->host, peer->port, pcache->tag, pcache->caid, pcache->sid, pcache->hash);
					pcache->peerid = peer->id;
					memcpy(pcache->cw, buf+13, 16);
					pcache->status = CACHE_STAT_DCW;
					//Send Reply to all peers
					p = cfg.cache.peer;
					while (p) 
						{
						if (p!=peer)
						if ( (peer->lastactivity+300000)>GetTickCount() ) // send only to active peers (5mn)
							send_reply(pcache,p);
						p = p->next;
						}
					}
				break;

			case TYPE_PINGREQ:
				if (received<13) break;
				// Send Reply
				buf[0] = TYPE_PINGRPL;
				sendtoip( recv_ip, (buf[11]<<8)|buf[12], buf, 9);
				// Check Peer
				peer = cfg.cache.peer;
				port = (buf[11]<<8)|buf[12];
				peerip = NULL;
				while (peer) {
					if (peer->ip==recv_ip) {
						if (peer->port==port) { // Peer is found
							peer->lastactivity = GetTickCount();
							// Check for extended reply
							int index = 13;
							while (received>index) {
								if ( (index+buf[index+1]+2)>received ) break;
								switch(buf[index]) {
									case 0x01:
										if (buf[index+1]<32) { memcpy(peer->program, buf+index+2, buf[index+1]); peer->program[buf[index+1]] = 0; }
										break;
									case 0x02:
										if (buf[index+1]<32) { memcpy(peer->version, buf+index+2, buf[index+1]); peer->version[buf[index+1]] = 0; }
										break;
								}
								index += 2+buf[index+1];
							}
							break;
						}
						peerip = peer;
					}
					peer = peer->next;
				}
				if (!peer && peerip) {
					peer = (struct cachepeer_data *)malloc( sizeof(struct cachepeer_data) );
					memset( peer, 0, sizeof(struct cachepeer_data) );
					strcpy( peer->host, peerip->host );
					peer->ip = recv_ip;
					peer->port = port;
					peer->lastactivity = GetTickCount();
					peer->next = cfg.cache.peer;
					peer->id = cfg.cachepeerid;
					cfg.cachepeerid++;
					cfg.cache.peer = peer;
					if (dbg) printf(" ++ new peer %s:%d\n", peer->host, peer->port);
				}
				break;

			case TYPE_PINGRPL:
				if (received<9) break;
				// Get Peer
				peerid = (buf[7]<<8) | buf[8];
				peer = cfg.cache.peer;
				while(peer) {
					if (peer->id==peerid) {
						peer->lastactivity = GetTickCount();
						if (dbg) printf(" < ping reply from %s:%d\n", peer->host, peer->port);
						peer->recvport = recv_port;
						peer->lastpingrecv = GetTickCount();
						peer->ping = peer->lastpingrecv-peer->lastpingsent;
						break;
					}
					peer = peer->next;
				}
				break;

			case TYPE_RESENDREQ:
				break; // nothing to do ( all cache data are already sent to peers )

			default:
				if (dbg) printf(" Cache: Unknown msg type\n"); // debughex(buf,received);
				break;
		}
	}
}


void CSP::cache_check_peers()
{
	unsigned int ticks = GetTickCount();
	struct cachepeer_data *peer = cfg.cache.peer;
	while (peer) {
		if ( (peer->ip)&&(peer->port) ) {

			if (peer->ping==0) { // inactive
				if ( (!peer->lastpingsent)||((peer->lastpingsent+15000)<ticks) ) { // send every 15s
					send_pingreq(peer);
					peer->lastpingsent = ticks;
					peer->lastpingrecv = 0;
				}
			}
			else {
				if ( (!peer->lastpingrecv)&&((peer->lastpingsent+1500)<ticks) ) { // Check for inactive peers
					peer->ping = 0;				
				}
				else if ( (peer->lastpingsent+75000)<ticks ) { // send every 75s
					send_pingreq(peer);
					peer->lastpingsent = ticks;
					peer->lastpingrecv = 0;
				}
			}

		}
		peer = peer->next;
	}
}




///CSP-Btk Ask Req
struct cache_data * CSP::BTK_to_CSP( struct cache_data *newdata )
{
struct cachepeer_data *p;
struct cache_data *pcache=NULL;

//printf(" - Query CSP Cache -> (%02x:%04x:%04x:%08x)\n",newdata->tag,newdata->caid,newdata->sid,newdata->hash);
pthread_mutex_lock( &mtxcsp );

pcache = cache_fetch( newdata );
if (pcache==NULL) 
	{
	pcache = cache_add( newdata );
	// Send to all Peers,
	p = cfg.cache.peer;
	while (p) {
		if ( (p->lastactivity+300000)>GetTickCount() ) // send only to active peers (5mn)
		send_request(pcache,p);
		p = p->next;
		}
	pthread_mutex_unlock( &mtxcsp );
	return NULL;
	}
//printf("Not Added Cache -> C:%04x   S:%04x",pcache->caid,pcache->sid);
if (pcache->status != CACHE_STAT_DCW) pcache=NULL;
pthread_mutex_unlock( &mtxcsp );
return pcache;
}


///////////////////////////////////////////////////////////////////////////////
// MAIN PROCESS
///////////////////////////////////////////////////////////////////////////////
void CSP::main()
{
fd_set readfds;
struct timeval timeout;
FD_ZERO(&readfds);
timeout.tv_usec = 0;
timeout.tv_sec = 1;
FD_SET( cfg.cache.handle, &readfds);
if ( select(FD_SETSIZE, &readfds, NULL, NULL,&timeout)>0 ) { 
	if ( FD_ISSET(cfg.cache.handle,&readfds) ) 
		{
		pthread_mutex_lock( &mtxcsp );
		cache_recvmsg();
		pthread_mutex_unlock( &mtxcsp );
		}
	}
// Check Peers Ping(KeepAlive)
cache_check_peers();
usleep(1000);
} 


CSP::CSP(char *server_path )
{
	ready=false;
	strcpy(config_file,server_path);
	strcat(config_file, "tracker.cfg");
	icachetab=0;
	dbg=0;
    	gettimeofday( &startime, NULL );
	// READ CFG
	if (read_config(&cfg)==-1) return;
	// Init DATA
	cfg.cache.handle = CreateServerSockUdp( cfg.cache.port );
	if (cfg.cache.handle<=0)  return;
	outsock = CreateClientSockUdp();
	if (outsock<=0)  return;
	memset( cachetab, 0, sizeof(cachetab) );
	//sleep(3); // wait for dns lookup 
	printf(" Entering main loop...\n");
	ready=true;
	pthread_mutex_init(&mtxcsp, NULL);
	//close(cfg.cache.handle);
	//return;
}



